import data_processing

import os
import re
from pathlib import Path

import openai
import pinecone
from pinecone import Pinecone as pinecone_client
import chromadb

from langchain_core.documents import Document

from langchain_openai import ChatOpenAI

from langchain_pinecone import Pinecone
from langchain_community.vectorstores import Chroma

from langchain.memory import ConversationBufferMemory

from langchain.retrievers.multi_query import MultiQueryRetriever
from langchain.retrievers.multi_vector import MultiVectorRetriever
from langchain.storage import LocalFileStore

from operator import itemgetter
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnableLambda, RunnablePassthrough
from langchain.schema import format_document
from langchain_core.messages import get_buffer_string

import data_processing
from prompts import CONDENSE_QUESTION_PROMPT, QA_PROMPT, DEFAULT_DOCUMENT_PROMPT, GENERATE_SIMILAR_QUESTIONS, GENERATE_SIMILAR_QUESTIONS_W_CONTEXT

# Class and functions
class QA_Model:
    """
    Represents a Question-Answering Model.

    Args:
        index_type (str): The type of document index.
        index_name (str): The name of the document index.
        query_model (str): The query model.
        llm (ChatOpenAI): The language model for generating responses.
        rag_type (str, optional): The type of RAG model.
        k (int, optional): The number of retriever results to consider.
        search_type (str, optional): The type of search to perform.
        fetch_k (int, optional): The number of documents to fetch from the retriever.
        temperature (int, optional): The temperature for response generation.
        local_db_path (str, optional): The path to the local database.
        reset_query_db (bool, optional): Whether to reset the query database.

    Attributes:
        index_type (str): The type of document index.
        index_name (str): The name of the document index.
        query_model (str): The query model.
        llm (ChatOpenAI): The language model for generating responses.
        rag_type (str): The type of RAG model.
        k (int): The number of retriever results to consider.
        search_type (str): The type of search to perform.
        fetch_k (int): The number of documents to fetch from the retriever.
        temperature (int): The temperature for response generation.
        local_db_path (str): The path to the local database.
        memory (ConversationBufferMemory): The memory for conversation history.
        result (dict): The result of the query.
        sources (list): The list of sources for the retrieved documents.
        ai_response (str): The response generated by the AI.
        conversational_qa_chain (Chain): The chain for conversational QA.
        doc_vectorstore (VectorStore): The vector store for document retrieval.
        query_vectorstore (VectorStore): The vector store for query retrieval. Only created with ChromaDB index_type.
        retriever (Retriever): The retriever for document retrieval.

    """
    def __init__(self, 
                 index_type:str,
                 index_name:str,
                 query_model:object,
                 llm:ChatOpenAI,
                 rag_type:str='Standard',
                 k:int=6,
                 search_type:str='similarity',
                 fetch_k:int=50,
                 temperature:int=0,
                 local_db_path:str='.',
                 reset_query_db:bool=False):
        """
        Initializes a new instance of the QA_Model class.

        Args:
            index_type (str): The type of document index.
            index_name (str): The name of the document index.
            query_model (object): The query model.
            llm (ChatOpenAI): The language model for generating responses.
            rag_type (str, optional): The type of RAG model.
            k (int, optional): The number of retriever results to consider.
            search_type (str, optional): The type of search to perform.
            fetch_k (int, optional): The number of documents to fetch from the retriever.
            temperature (int, optional): The temperature for response generation.
            local_db_path (str, optional): The path to the local database.
            reset_query_db (bool, optional): Whether to reset the query database.

        """
        self.index_type=index_type
        self.index_name=index_name
        self.query_model=query_model
        self.llm=llm
        self.rag_type=rag_type
        self.k=k
        self.search_type=search_type
        self.fetch_k=fetch_k
        self.temperature=temperature
        self.local_db_path=local_db_path

        self.memory=None
        self.result=None
        self.sources=None
        self.ai_response=None
        self.conversational_qa_chain=None

        self.doc_vectorstore=None
        self.query_vectorstore=None
        self.retriever=None

        # Define retriever search parameters
        search_kwargs = _process_retriever_args(self.search_type,
                                                self.k,
                                                self.fetch_k)

        # Read in from the vector database
        self.doc_vectorstore=data_processing.initialize_database(self.index_type,
                                                             self.index_name,
                                                             self.query_model,
                                                             self.rag_type,
                                                             local_db_path=self.local_db_path,
                                                             clear=False,
                                                             init_ragatouille=False)  
        
        # Iniialize a database to capture queries in a temp database
        if self.index_type=='ChromaDB':
            self.query_vectorstore=data_processing.initialize_database(self.index_type,
                                                                 self.index_name+'-queries',
                                                                 self.query_model,
                                                                 'Standard',    # Regardless of doc_vectorstore, query_vectorstore is always Standard
                                                                 local_db_path=self.local_db_path,
                                                                 clear=reset_query_db)

        # Initialize retriever
        if self.rag_type=='Standard':  
            if self.index_type=='ChromaDB' or self.index_type=='Pinecone':
                self.retriever=self.doc_vectorstore.as_retriever(search_type=self.search_type,
                                                                 search_kwargs=search_kwargs)
            elif self.index_type=='RAGatouille':
                self.retriever=self.doc_vectorstore.as_langchain_retriever(k=search_kwargs['k'])  
        elif self.rag_type=='Parent-Child' or self.rag_type=='Summary':
            self.lfs = LocalFileStore(Path(self.local_db_path).resolve() / 'local_file_store' / self.index_name)
            self.retriever = MultiVectorRetriever(
                                vectorstore=self.doc_vectorstore,
                                byte_store=self.lfs,
                                id_key="doc_id",
                                search_type=self.search_type,
                                search_kwargs=search_kwargs)
        else:
            raise NotImplementedError

        # Intialize memory
        self.memory = ConversationBufferMemory(
                        return_messages=True, output_key='answer', input_key='question')

        # Assemble main chain
        self.conversational_qa_chain=_define_qa_chain(self.llm,
                                                      self.retriever,
                                                      self.memory)
    def query_docs(self,query): 
        """
        Executes a query and retrieves the relevant documents.

        Args:
            query (str): The query string.

        Returns:
            None
        """       

        # Retrieve memory, invoke chain
        self.memory.load_memory_variables({})

        # TODO add error checking for max token length of a given model.
        self.result = self.conversational_qa_chain.invoke({'question': query})

        # Add sources to response
        self.sources = '\n'.join(str(data.metadata) for data in self.result['references'])
        if self.llm.__class__.__name__=='ChatOpenAI':
            self.ai_response = self.result['answer'].content + '\n\nSources:\n' + self.sources
        else:
            raise NotImplementedError
        self.memory.save_context({'question': query}, {'answer': self.ai_response})

        # If ChromaDB type, upsert query into query database
        if self.index_type=='ChromaDB':
            self.query_vectorstore.add_documents([_question_as_doc(query, self.result)])
        
    def update_model(self,
                     llm:ChatOpenAI):
        """
        Updates with a new LLM.

        Args:
            llm (ChatOpenAI): The language model for generating responses.

        Returns:
            None
        """
        # TODO add in updated retrieval parameters
        self.llm=llm

        # Update conversational retrieval chain
        self.conversational_qa_chain=_define_qa_chain(self.llm,
                                                      self.retriever,
                                                      self.memory)
    def generate_alternative_questions(self,
                                       prompt:str):
        """
        Generates alternative questions based on a prompt.

        Args:
            prompt (str): The prompt for generating alternative questions.

        Returns:
            str: The generated alternative questions.
        """
        if self.ai_response:
            prompt_template=GENERATE_SIMILAR_QUESTIONS_W_CONTEXT
            invoke_dict={'question':prompt,'context':self.ai_response}
        else:
            prompt_template=GENERATE_SIMILAR_QUESTIONS
            invoke_dict={'question':prompt}
        
        chain = (
                prompt_template
                | self.llm
                | StrOutputParser()
            )
        
        alternative_questions=chain.invoke(invoke_dict)
        return alternative_questions

# Internal functions
def _combine_documents(docs, 
                       document_prompt=DEFAULT_DOCUMENT_PROMPT, 
                       document_separator='\n\n'):
    """Combines a list of documents into a single string.

    Args:
        docs (list): A list of documents to be combined.
        document_prompt (str, optional): The prompt to be added before each document.
        document_separator (str, optional): The separator to be added between each document.

    Returns:
        str: The combined string of all the documents.
    """
    doc_strings = [format_document(doc, document_prompt) for doc in docs]
    return document_separator.join(doc_strings)

def _define_qa_chain(llm,
                     retriever,
                     memory):
    """Defines the conversational QA chain.

    Args:
        llm: The language model component.
        retriever: The document retriever component.
        memory: The memory component.

    Returns:
        The conversational QA chain.

    Raises:
        None.
    """
    # This adds a 'memory' key to the input object
    loaded_memory = RunnablePassthrough.assign(
                        chat_history=RunnableLambda(memory.load_memory_variables) 
                        | itemgetter('history'))  
    
    # Assemble main chain
    standalone_question = {
        'standalone_question': {
            'question': lambda x: x['question'],
            'chat_history': lambda x: get_buffer_string(x['chat_history'])}
        | CONDENSE_QUESTION_PROMPT
        | llm
        | StrOutputParser()}
    
    retrieved_documents = {
        'source_documents': itemgetter('standalone_question') 
                            | retriever,
        'question': lambda x: x['standalone_question']}
    
    # Now we construct the inputs for the final prompt
    final_inputs = {
        'context': lambda x: _combine_documents(x['source_documents']),
        'question': itemgetter('question')}
    
    # And finally, we do the part that returns the answers
    answer = {
        'answer': final_inputs 
                    | QA_PROMPT 
                    | llm,
        'references': itemgetter('source_documents')}
    
    conversational_qa_chain = loaded_memory | standalone_question | retrieved_documents | answer
    return conversational_qa_chain
def _process_retriever_args(search_type='similarity',
                            k=6,
                            fetch_k=50):
    """
    Process the retriever arguments.

    Args:
        search_type (str, optional): The type of search.
        k (int, optional): The number of documents to retrieve.
        fetch_k (int, optional): The number of documents to fetch.

    Returns:
        dict: The search arguments for the retriever.
    """
    # TODO add functionality for filter if required
    # # Implement filter
    # if filter_arg:
    #     filter_list = list(set(item['source'] for item in sources[-1]))
    #     filter_items=[]
    #     for item in filter_list:
    #         filter_item={'source': item}
    #         filter_items.append(filter_item)
    #     filter={'$or':filter_items}
    # else:
    #     filter=None

    # Implement filtering and number of documents to return
    if search_type=='mmr':
        search_kwargs={'k':k,'fetch_k':fetch_k} # See as_retriever docs for parameters
    else:
        search_kwargs={'k':k} # See as_retriever docs for parameters
    
    return search_kwargs
def _question_as_doc(question: str, rag_answer: dict):
    """
    Creates a Document object based on the given question and RAG answer.

    Args:
        question (str): The question to be included in the page content of the Document.
        rag_answer (dict): The RAG answer containing the references and answer content.

    Returns:
        Document: The created Document object.

    """
    sources = [data.metadata for data in rag_answer['references']]

    return Document(
        page_content=question,
        metadata={
            "answer": rag_answer['answer'].content,
            "sources": ",".join(map(data_processing._stable_hash_meta, sources)),
        },
    )


import os, sys
import json
from pathlib import Path
import pickle

from ragas.testset import TestsetGenerator
from ragas import RunConfig
from dotenv import load_dotenv,find_dotenv
import chromadb
from chromadb import PersistentClient
from pinecone import Pinecone as pinecone_client, ServerlessSpec
from langchain_pinecone import PineconeVectorStore
from langchain_community.vectorstores import Chroma
from langchain_openai import ChatOpenAI, OpenAIEmbeddings
from langchain_voyageai import VoyageAIEmbeddings
from langchain_community.embeddings import HuggingFaceInferenceAPIEmbeddings
from langchain_core.documents import Document
import pandas as pd

from ragas import evaluate
from ragas.metrics import answer_correctness, faithfulness, context_recall
from datasets import Dataset

from renumics import spotlight
from renumics.spotlight import Embedding
import pandas as pd

from umap import UMAP
import numpy as np

# Import local packages
sys.path.append('../src/aerospace_chatbot')
import queries
import data_processing

# TODO after tests are written, write docstrings

# Set environment variables with .env
load_dotenv(find_dotenv(), override=True)

def write_dict_to_file(data_dict, filename):
    """write a dictionary as a json line to a file - allowing for appending"""
    with open(filename, "a") as f:
        f.write(json.dumps(data_dict) + "\n")

def read_dicts_from_file(filename):   
    # For this to work, each dict must be on a separate line!
    # Initialize an empty list to hold the JSON data
    json_data = []

    # Open the file and read line by line
    with open(filename, 'r') as file:
        for line in file:
            # Parse each line as JSON and append to the list
            json_data.append(json.loads(line))
    return json_data

def add_cached_columns_from_file(df, file_name, merge_on, columns,filter=None):
    """
    Read a file with cached list of dicts data and write it to a dataframe.
    filter is a dict with keys which are column headers and values which are the filters to apply to the columns.
    """
    if Path(file_name).exists():
        df_in=pd.DataFrame(list(read_dicts_from_file(file_name)))
        # Filter the cached file by filter if it exists.
        if filter is not None:  
            # filtered_data = df_in.copy()
            for column, value in filter.items():
                df_in = df_in[df_in[column] == value]

        cached_data = (
            df_in
            .drop_duplicates(
                subset=[merge_on],
            )[columns + [merge_on]]
            .dropna()
            .reset_index(drop=True)
        )
        return df.merge(
            cached_data,
            on=merge_on,
            how="left",
        ).replace({float('nan'): None}).reset_index(drop=True)
    else:
        # Create a copy of the DataFrame
        df_out = df.copy()

        # Add the new columns with the names specified in the 'columns' list
        if isinstance(columns, str):    # Handle where it's a single value, not an array of columns
            columns = [columns]
        for column in columns:
            df_out[column] = None

        # Reorder the columns to place the new columns at the end
        columns = list(df_out.columns)
        for column in columns:
            if column not in columns:
                columns.append(column)
        df_out = df_out[columns]

        return df_out
    
def lcdoc_export(index_type, index, query_model, export_pickle=False):
    if index_type=="ChromaDB":
        # Inspect the first db, save for synthetic test dataset
        all_docs = index.get(include=["metadatas", "documents", "embeddings"])
        lcdocs = [Document(page_content=doc, metadata=metadata) 
                for doc, metadata in zip(all_docs['documents'], all_docs['metadatas'])]
        
        # Format docs into dataframe
        all_docs = index.get(include=["metadatas", "documents", "embeddings"])
        df_docs = pd.DataFrame(
            {
                "id": [data_processing._stable_hash_meta(metadata) for metadata in all_docs["metadatas"]],
                "source": [metadata.get("source") for metadata in all_docs["metadatas"]],
                "page": [metadata.get("page", -1) for metadata in all_docs["metadatas"]],
                "document": all_docs["documents"],
                "embedding": all_docs["embeddings"],
            }
        )
        if export_pickle:
            df_temp=data_processing.archive_db('ChromaDB',index.name,query_model,export_pickle=True)
        
    elif index_type=="Pinecone":
        ids=[]
        for id in index.list():
            ids.extend(id)

        docs=[]
        df_docs = pd.DataFrame()
        chunk_size=200  # Tune to whatever doesn't error out, 200 won't for serverless
        for i in range(0, len(ids), chunk_size):
            print(f"Fetching {i} to {i+chunk_size}")
            vector=index.fetch(ids[i:i+chunk_size])['vectors']
            vector_data = []
            for key, value in vector.items():
                vector_data.append(value)
            docs.extend(vector_data)

            df_doc_temp = pd.DataFrame()
            df_doc_temp["id"]= [vector_elm["id"] for vector_elm in vector_data]
            df_doc_temp["source"]= [vector_elm["metadata"]["source"] for vector_elm in vector_data]
            df_doc_temp["page"]= [vector_elm["metadata"]["page"] for vector_elm in vector_data]
            df_doc_temp["document"]= [vector_elm["metadata"]["page_content"] for vector_elm in vector_data]
            df_doc_temp["embedding"]= [vector_elm["values"] for vector_elm in vector_data]
            df_docs = pd.concat([df_docs, df_doc_temp])

        lcdocs = []
        for data in docs:
            data=data['metadata']
            lcdocs.append(Document(page_content=data['page_content'],
                                metadata={'page':data['page'],'source':data['source']}))
        if export_pickle:
            df_temp=data_processing.archive_db('Pinecone',db['index_name'],query_model,export_pickle=True)
    
    return df_docs, lcdocs
        

def synthetic_dataset_loop(lcdocs,eval_size,n_questions,fname):
    """ 
    Check if testset.csv exists, use, or generate the synthetic dataset. If it doesn't exist, just loop through everything.
    If a partial export exists, it will loop through the remaining indices.
    If a few entries are missing it will loop through that full index.
    """
    if os.path.exists(fname):
        # Import testset.csv into a DataFrame
        df_testset = pd.read_csv(fname)

        full_index_list = list(range(0, len(lcdocs), eval_size))
        index_values = df_testset['index'].unique()
        index_loop = [] # Indices to index over

        for index_value in index_values:
            if len(df_testset[df_testset['index'] == index_value]) != n_questions:
                index_loop.append(index_value)

        index_loop += list(set([index for index in full_index_list if index not in index_values]))
        index_loop = sorted(index_loop)
    else:
        df_testset = pd.DataFrame()  # Initialize an empty DataFrame
        index_loop = list(range(0, len(lcdocs), eval_size))
    return df_testset, index_loop

def generate_testset(lcdocs,generator,eval_size,n_questions,fname,run_config):
    """Loop through eval_size chunks of the dataset to generate the testset. Export to csv along the way."""
    # Initialize the DataFrame to store the testset, determine what parts of the dataset need to be looped over
    df_testset, index_loop = synthetic_dataset_loop(lcdocs,eval_size,n_questions,fname)
    print(f"Index loop: {index_loop}")

    for i in index_loop:
        print(f"Processing index {i} to {i+eval_size}...")
        lcdocs_eval = lcdocs[i:i+eval_size]

        testset = generator.generate_with_langchain_docs(lcdocs_eval, 
                                                        test_size=n_questions,
                                                        with_debugging_logs=True,
                                                        is_async=False, # Avoid rate limit issues
                                                        run_config=run_config,
                                                        raise_exceptions=False)

        df_testset_new = testset.to_pandas()
        df_testset_new['index'] = i
        df_testset_new['eval_size'] = eval_size
        columns = ['index', 'eval_size'] + [col for col in df_testset_new.columns if col not in ['index', 'eval_size']]
        df_testset_new = df_testset_new[columns]

        # Export the testset to a csv file with the index
        if os.path.exists(fname):
            df_testset_new.to_csv(fname, mode='a', header=False, index=False)
        else:
            df_testset_new.to_csv(fname, index=False)
        df_testset = pd.concat([df_testset, df_testset_new])
    return df_testset

def rag_responses(index_type, index_name, query_model, llm, QA_model_params, df_qa, df_docs, testset_name):
    df_qa_out=df_qa.copy()

    # RAGatouille has a different structure than the other models
    if index_type == "RAGatouille":
        query_model_name=query_model.model.checkpoint
    else:
        query_model_name=query_model.model

    # Load cached version of rag responses, filter the responses by the model and parameters being evaluated
    df_qa_out = add_cached_columns_from_file(
        df_qa_out, 
        os.path.join('output',f'rag_response_cache_{testset_name}.jsonl'), "question", 
        ["answer", "source_documents", "answer_by", "query_model", "qa_model_params","index_type","index_name"],
        filter={"answer_by": llm.model_name,
                "query_model": query_model_name, 
                "qa_model_params": QA_model_params,
                "index_type": index_type,
                "index_name": index_name}
    )

    # Generate responses using RAG with input parameters
    for i, row in df_qa_out.iterrows():
        # Loop through each row
        if (row['answer_by'] != llm.model_name) or \
           (row['query_model'] != query_model_name) or \
           (row['qa_model_params'] != str(QA_model_params)):    # Check if the model and parameters are the same

            if row['answer'] is None or pd.isnull(row['answer']) or row['answer']=='':  # Check if the answer is empty
                print(f"Processing question {i+1}/{len(df_qa_out)}")


                # Use the QA model to query the documents
                qa_obj=queries.QA_Model(index_type,
                                index_name,
                                query_model,
                                llm,
                                **QA_model_params)
                qa_obj.query_docs(row['question'])
                response=qa_obj.result

                df_qa_out.loc[df_qa_out.index[i], "answer"] = response['answer'].content

                ids=[data_processing._stable_hash_meta(source_document.metadata)
                    for source_document in response['references']]
                df_qa_out.loc[df_qa_out.index[i], "source_documents"] = ', '.join(ids)
                # df_qa_out.loc[df_qa_out.index[i], "source_documents"] = ids

                df_qa_out.loc[df_qa_out.index[i], "answer_by"] = llm.model_name
                df_qa_out.loc[df_qa_out.index[i], "query_model"] = query_model_name
                df_qa_out.loc[df_qa_out.index[i], "qa_model_params"] = str(QA_model_params)
                df_qa_out.loc[df_qa_out.index[i], "index_type"] = index_type
                df_qa_out.loc[df_qa_out.index[i], "index_name"] = index_name

                # Save the response to cache file
                response_dict = {
                    "question_id": row['question_id'],
                    "question": row['question'],
                    "answer": response['answer'].content,
                    "source_documents": ids,
                    "answer_by": llm.model_name,
                    "query_model": query_model_name,
                    "qa_model_params": QA_model_params,
                    "index_type": index_type,
                    "index_name": index_name
                }
                write_dict_to_file(response_dict, os.path.join('output',f'rag_response_cache_{testset_name}.jsonl'))

    # Get the context documents content for each question
    source_documents_list = []
    for cell in df_qa_out['source_documents']:
        if isinstance(cell, str):
            cell_list = cell.split(', ')
        else:
            cell_list = cell
        context=[]
        for cell in cell_list:
            context.append(df_docs[df_docs["id"] == cell]["document"].values[0])
        source_documents_list.append(context)
    df_qa_out["contexts"]=source_documents_list

    # Addtionaly get embeddings for questions
    if index_type != "RAGatouille":
        if not Path(os.path.join('output',f'question_embeddings_{testset_name}.pickle')).exists():
            question_embeddings = [
                query_model.embed_query(question)
                for question in df_qa_out["question"]
        ]
        df_qa_out["embedding"] = question_embeddings
    else:
        # TODO add RAGatouille encodings
        df_qa_out["embedding"] = None

    return df_qa_out

def eval_rag(df_qa, eval_criterias, testset_name):
    df_qa_out=df_qa.copy()

    # Add answer correctness column, fill in if it exists
    df_qa_out = add_cached_columns_from_file(
        df_qa_out, 
        os.path.join('output',f'ragas_result_cache_{testset_name}.jsonl'), 
        "question", 
        eval_criterias
    )

    # Sometimes ground_truth does not provide a response. Just filter those out.
    df_qa_out = df_qa_out[df_qa_out['ground_truth'].apply(lambda x: isinstance(x, str))]

    # Evaluate the answer correctness if not already done
    fields = ["question", "answer", "contexts", "ground_truth"]

    for i, row in df_qa_out.iterrows():
        print(i, row["question"])
        response_dict={}
        response_dict["question"]=row["question"]

        if any(row[eval_criteria] is None for eval_criteria in eval_criterias):
            for eval_criteria in eval_criterias:
                print(eval_criteria)
                if eval_criteria=="answer_correctness":
                    eval_obj=answer_correctness
                elif eval_criteria=="faithfulness":
                    eval_obj=faithfulness
                elif eval_criteria=="context_recall":
                    eval_obj=context_recall

                if row[eval_criteria] is None or pd.isnull(row[eval_criteria]):
                    evaluation_result = evaluate(
                        Dataset.from_pandas(df_qa_out.iloc[i : i + 1][fields]),
                        [eval_obj],
                    )
                    df_qa_out.loc[i,eval_criteria] = evaluation_result[
                        eval_criteria
                    ]
                    
                response_dict[eval_criteria]=evaluation_result[eval_criteria]
            write_dict_to_file(response_dict, os.path.join('output',f'ragas_result_cache_{testset_name}.jsonl'))

    df_qa=df_qa_out

    return df_qa_out

def data_viz_prep(index_name,df_qa_eval,df_docs):
    """This section adds a column to df_documents containing the ids of the questions that used the document as source. """

    # add the infos about questions using each document to the documents dataframe
    # Explode 'source_documents' so each document ID is in its own row alongside the question ID
    df_questions_exploded = df_qa_eval.explode("source_documents")

    # Group by exploded 'source_documents' (document IDs) and aggregate
    agg = (
        df_questions_exploded.groupby("source_documents")
        .agg(
            num_questions=("id", "count"),  # Count of questions referencing the document
            question_ids=(
                "id",
                lambda x: list(x),
            ),  # List of question IDs referencing the document
        )
        .reset_index()
        .rename(columns={"source_documents": "id"})
    )

    # Merge the aggregated information back into df_documents
    df_documents_agg = pd.merge(df_docs, agg, on="id", how="left")

    # Use apply to replace NaN values with empty lists for 'question_ids'
    df_documents_agg["question_ids"] = df_documents_agg["question_ids"].apply(
        lambda x: x if isinstance(x, list) else []
    )
    # Replace NaN values in 'num_questions' with 0
    df_documents_agg["num_questions"] = df_documents_agg["num_questions"].fillna(0)

    # Concatenate the two dataframes
    df_visualize = pd.concat([df_qa_eval, df_documents_agg], axis=0)

    df_questions = df_visualize[~df_visualize["question"].isna()]
    umap = UMAP(n_neighbors=20, min_dist=0.15, metric="cosine", random_state=42).fit(
        df_questions["embedding"].values.tolist()
    )
    # umap_questions = umap.transform(df_visualize["embedding"].values.tolist())


    df_without_questions = df_visualize[df_visualize["question"].isna()]
    umap = UMAP(n_neighbors=20, min_dist=0.15, metric="cosine", random_state=42).fit(
        df_without_questions["embedding"].values.tolist()
    )
    umap_docs = umap.transform(df_visualize["embedding"].values.tolist())
    df_visualize["umap_docs"] = umap_docs.tolist()

    umap = UMAP(n_neighbors=20, min_dist=0.15, metric="cosine", random_state=42).fit(
        df_visualize["embedding"].values.tolist()
    )
    umap_all = umap.transform(df_visualize["embedding"].values.tolist())
    df_visualize["umap"] = umap_all.tolist()


    # find the nearet question (by embedding) for each document
    question_embeddings = np.array(df_visualize[df_visualize["question"].notna()]["embedding"].tolist())

    df_visualize["nearest_question_dist"] = [  # brute force, could be optimized using ChromaDB
        np.min([np.linalg.norm(np.array(doc_emb) - question_embeddings, axis=1)])
        for doc_emb in df_visualize["embedding"].values
    ]

    # write the dataframe to parquet for later use
    df_visualize.to_parquet(f'df_{index_name}.parquet')

    return df_visualize


from prompts import SUMMARIZE_TEXT

import os, re, shutil
import hashlib
from pathlib import Path
from typing import List, Union
import pickle

from tenacity import retry, stop_after_attempt, wait_exponential

from pinecone import Pinecone as pinecone_client, ServerlessSpec

import chromadb
from chromadb import PersistentClient

import json, jsonlines

import streamlit as st

from langchain_pinecone import PineconeVectorStore
from langchain_community.vectorstores import Chroma

from langchain.text_splitter import RecursiveCharacterTextSplitter

from langchain.retrievers.multi_vector import MultiVectorRetriever
from langchain.storage import LocalFileStore

from langchain_openai import OpenAIEmbeddings
from langchain_voyageai import VoyageAIEmbeddings
from langchain_community.embeddings import HuggingFaceInferenceAPIEmbeddings

from langchain_community.document_loaders import PyPDFLoader

from langchain_core.documents import Document
from langchain_core.output_parsers import StrOutputParser

from ragatouille import RAGPretrainedModel

from renumics import spotlight
from renumics.spotlight import dtypes as spotlight_dtypes

import pandas as pd
import numpy as np
from sklearn.cluster import KMeans

from datasets import Dataset

import admin
from prompts import CLUSTER_LABEL
import time

def load_docs(index_type:str,
              docs:List[str],
              query_model:object,
              rag_type:str='Standard',
              index_name:str=None,
              n_merge_pages:int=None,
              chunk_method:str='character_recursive',
              chunk_size:int=500,
              chunk_overlap:int=0,
              clear:bool=False,
              file_out:str=None,
              batch_size:int=50,
              local_db_path:str='.',
              llm=None,
              show_progress:bool=False):
    """
    Loads documents into the specified index.

    Args:
        index_type (str): The type of index to use.
        docs: The documents to load.
        query_model (object): The query model to use.
        rag_type (str, optional): The type of RAG (Retrieval-Augmented Generation) to use.
        index_name (str, optional): The name of the index.
        n_merge_pages (int, optional): Number of pages to to merge when loading.
        chunk_method (str, optional): The method to chunk the documents.
        chunk_size (int, optional): The size of each chunk.
        chunk_overlap (int, optional): The overlap between chunks.
        clear (bool, optional): Whether to clear the index before loading new documents.
        file_out (str, optional): The output file path.
        batch_size (int, optional): The batch size for upserting documents.
        local_db_path (str, optional): The local database path.
        llm (optional): The language model to use.
        show_progress (bool, optional): Whether to show progress during the loading process.

    Returns:
        vectorstore: The updated vectorstore.
    """
    # Check for illegal things
    if not clear and (rag_type == 'Parent-Child' or rag_type == 'Summary'):
        raise ValueError('Parent-Child databases must be cleared before loading new documents.')

    # Chunk docs
    chunker=chunk_docs(docs,
                       rag_type=rag_type,
                       n_merge_pages=n_merge_pages,
                       chunk_method=chunk_method,
                       chunk_size=chunk_size,
                       chunk_overlap=chunk_overlap,
                       file_out=file_out,
                       llm=llm,
                       show_progress=show_progress)

    # Initialize client an upsert docs
    vectorstore = initialize_database(index_type, 
                                      index_name, 
                                      query_model,
                                      rag_type=rag_type,
                                      clear=clear, 
                                      local_db_path=local_db_path,
                                      init_ragatouille=True,
                                      show_progress=show_progress)
    vectorstore, _ = upsert_docs(index_type,
                                 index_name,
                                 vectorstore,
                                 chunker,
                                 batch_size=batch_size,
                                 show_progress=show_progress,
                                 local_db_path=local_db_path)
    return vectorstore
def chunk_docs(docs: List[str],
               rag_type:str='Standard',
               chunk_method:str='character_recursive',
               file_out:str=None,
               n_merge_pages:int=None,
               chunk_size:int=400,
               chunk_overlap:int=0,
               k_child:int=4,
               llm=None,
               show_progress:bool=False):
    """
    Chunk the given list of documents into smaller chunks based on the specified parameters.

    Args:
        docs (List[str]): The list of document paths to be chunked.
        rag_type (str, optional): The type of chunking method to be used.
        chunk_method (str, optional): The method of chunking to be used. None will take whole PDF pages as documents.
        file_out (str, optional): The output file path to save the chunked documents.
        n_merge_pages (int, optional): Number of pages to to merge when loading.
        chunk_size (int, optional): The size of each chunk in tokens. Defaults to 500. Only used if chunk_method is not None.
        chunk_overlap (int, optional): The overlap between chunks in tokens. Defaults to 0. Only used if chunk_method is not None.
        k_child (int, optional): The number of child chunks to split from parnet chunks for 'Parent-Child' rag_type.
        llm (None, optional): The language model to be used for generating summaries.
        show_progress (bool, optional): Whether to show the progress bar during chunking.

    Returns:
        dict: A dictionary containing the chunking results based on the specified rag_type.
    """
    if show_progress:
        progress_text = 'Reading documents...'
        my_bar = st.progress(0, text=progress_text)
    pages=[]
    chunks=[]
    
    # Parse doc pages
    for i, doc in enumerate(docs):
        # Show and update the progress bar
        if show_progress:
            progress_percentage = i / len(docs)
            my_bar.progress(progress_percentage, text=f'Reading documents...{doc}...{progress_percentage*100:.2f}%')
        
        # Load the document
        loader = PyPDFLoader(doc)
        doc_page_data = loader.load()

        # Clean up page info, update some metadata
        doc_pages=[]
        for doc_page in doc_page_data:
            doc_page=_sanitize_raw_page_data(doc_page)
            if doc_page is not None:
                doc_pages.append(doc_page)

        # Merge pages if option is selected
        if n_merge_pages:
            for i in range(0, len(doc_pages), n_merge_pages):
                group = doc_pages[i:i+n_merge_pages]
                group_page_content=' '.join([doc.page_content for doc in group])
                group_metadata = {'page': str([doc.metadata['page'] for doc in group]), 
                                  'source': str([doc.metadata['source'] for doc in group])}
                merged_doc = Document(page_content=group_page_content, metadata=group_metadata)
                pages.append(merged_doc)
        else:
            pages.extend(doc_pages)
    
    # Process pages
    if rag_type=='Standard': 
        if chunk_method=='character_recursive':
            text_splitter = RecursiveCharacterTextSplitter(chunk_size=chunk_size, 
                                                           chunk_overlap=chunk_overlap,
                                                           add_start_index=True)
            page_chunks = text_splitter.split_documents(pages)
            for i, chunk in enumerate(page_chunks):
                if show_progress:
                    progress_percentage = i / len(page_chunks)
                    my_bar.progress(progress_percentage, text=f'Chunking documents...{progress_percentage*100:.2f}%')
                chunks.append(chunk)    # Not sanitized because the page already was
        elif chunk_method=='None':
            text_splitter = None
            chunks = pages  # No chunking, take whole pages as documents
        else:
            raise NotImplementedError
        
        if file_out:
            # Write to a jsonl file, save it.
            chunk_ids = [_stable_hash_meta(chunk.metadata) for chunk in chunks]   # add ID which is the hash of metadata
            with jsonlines.open(file_out, mode='w') as writer:
                for doc, chunk_id in zip(chunks, chunk_ids): 
                    doc_out=doc.dict()
                    doc_out['chunk_id'] = chunk_id  # Add chunk_id to the jsonl file
                    writer.write(doc_out)
        if show_progress:
            my_bar.empty()
        return {'rag':'Standard',
                'pages':pages,
                'chunks':chunks, 
                'splitters':text_splitter}
    elif rag_type=='Parent-Child': 
        if chunk_method=='character_recursive':
            # Settings apply to parent splitter. k_child divides parent into smaller sizes.
            parent_splitter=RecursiveCharacterTextSplitter(chunk_size=chunk_size, 
                                                           chunk_overlap=chunk_overlap,
                                                           add_start_index=True)    # Without add_start_index, will not be a unique id
            parent_chunks = parent_splitter.split_documents(pages)
        elif chunk_method=='None':
            parent_splitter = None
            parent_chunks = pages  # No chunking, take whole pages as documents
            # raise ValueError("You must specify a chunk_method with rag_type=Parent-Child.")
        else:
            raise NotImplementedError
        
        # Assign parent doc ids
        doc_ids = [str(_stable_hash_meta(parent_chunk.metadata)) for parent_chunk in parent_chunks]
        
        # Split up child chunks
        id_key = "doc_id"
        chunks = []
        for i, doc in enumerate(parent_chunks):
            _id = doc_ids[i]

            if chunk_method=='character_recursive':
                child_splitter=RecursiveCharacterTextSplitter(chunk_size=chunk_size/k_child, 
                                                chunk_overlap=chunk_overlap,
                                                add_start_index=True)
            elif chunk_method=='None':
                i_chunk_size=len(doc.page_content)/k_child
                child_splitter=RecursiveCharacterTextSplitter(chunk_size=i_chunk_size, 
                                                chunk_overlap=0,
                                                add_start_index=True)

            _chunks = child_splitter.split_documents([doc])
            for _doc in _chunks:
                _doc.metadata[id_key] = _id
            chunks.extend(_chunks)

        if show_progress:
            my_bar.empty()
        return {'rag':'Parent-Child',
                'pages':{'doc_ids':doc_ids,'parent_chunks':parent_chunks},
                'chunks':chunks,
                'splitters':{'parent_splitter':parent_splitter,'child_splitter':child_splitter}}
    elif rag_type == 'Summary':
        if show_progress:
            my_bar.empty()
            my_bar = st.progress(0, text='Generating summaries...')

        id_key = "doc_id"
        doc_ids = [str(_stable_hash_meta(page.metadata)) for page in pages]
        chain = (
            {"doc": lambda x: x.page_content}
            | SUMMARIZE_TEXT
            | llm
            | StrOutputParser()
        )

        summaries = []
        batch_size_submit = 10  # Set the batch size
        for i in range(0, len(pages), batch_size_submit):
            batch_pages = pages[i:i+batch_size_submit]  # Get the batch of pages
            summary = chain.batch(batch_pages, config={"max_concurrency": batch_size_submit})
            summaries.extend(summary)
            if show_progress:
                progress_percentage = i / len(pages)
                my_bar.progress(progress_percentage, text=f'Generating summaries...{progress_percentage*100:.2f}%')

        summary_docs = [
            Document(page_content=s, metadata={id_key: doc_ids[i]})
            for i, s in enumerate(summaries)
        ]
        if show_progress:
            my_bar.empty()
        return {'rag':'Summary',
                'pages':{'doc_ids':doc_ids,'docs':pages},
                'summaries':summary_docs,
                'llm':llm}
    else:
        raise NotImplementedError
def initialize_database(index_type: str, 
                        index_name: str, 
                        query_model: object,
                        rag_type: str,
                        local_db_path: str = None, 
                        clear: bool = False,
                        init_ragatouille: bool = False,
                        show_progress: bool = False):
    """Initializes the database based on the specified parameters.

    Args:
        index_type (str): The type of index to use (e.g., "Pinecone", "ChromaDB", "RAGatouille").
        index_name (str): The name of the index.
        query_model (object): The query model to use.
        rag_type (str): The type of RAG model to use.
        local_db_path (str, optional): The path to the local database.
        clear (bool, optional): Whether to clear the index.
        init_ragatouille (bool, optional): Whether to initialize the RAGatouille model.
        show_progress (bool, optional): Whether to show the progress bar.

    Returns:
        vectorstore: The initialized vector store.

    Raises:
        NotImplementedError: If the specified index type is not implemented.
    """

    if show_progress:
        progress_text = "Database initialization..."
        my_bar = st.progress(0, text=progress_text)

    if index_type == "Pinecone":

        if clear:
            delete_index(index_type, index_name, rag_type, local_db_path=local_db_path)
        pc = pinecone_client(api_key=os.getenv('PINECONE_API_KEY'))
        
        try:
            pc.describe_index(index_name)
        except:
            pc.create_index(index_name,
                            dimension=_embedding_size(query_model),
                            spec=ServerlessSpec(cloud="aws", region="us-east-1"))
        
        index = pc.Index(index_name)
        vectorstore=PineconeVectorStore(index,
                                        index_name=index_name, 
                                        embedding=query_model,
                                        text_key='page_content',
                                        pinecone_api_key=os.getenv('PINECONE_API_KEY'))
        if show_progress:
            progress_percentage = 1
            my_bar.progress(progress_percentage, text=f'{progress_text}{progress_percentage*100:.2f}%')
    elif index_type == "ChromaDB":
        # TODO add in collection metadata
        if clear:
            delete_index(index_type, index_name, rag_type, local_db_path=local_db_path)
        persistent_client = chromadb.PersistentClient(path=os.path.join(local_db_path,'chromadb'))            
        vectorstore = Chroma(client=persistent_client,
                                collection_name=index_name,
                                embedding_function=query_model) 
        if show_progress:
            progress_percentage = 1
            my_bar.progress(progress_percentage, text=f'{progress_text}{progress_percentage*100:.2f}%')   
    elif index_type == "RAGatouille":
        if clear:
            delete_index(index_type, index_name, rag_type, local_db_path=local_db_path)
        if init_ragatouille:    
            # Used if the index is not already set, initializes root folder and embedding model
            vectorstore = query_model
        else:   
            # Used if the index is already set, loads the index directly
            vectorstore = RAGPretrainedModel.from_index(index_path=os.path.join(local_db_path,
                                                                                '.ragatouille/colbert/indexes',
                                                                                index_name))
        if show_progress:
            progress_percentage = 1
            my_bar.progress(progress_percentage, text=f'{progress_text}{progress_percentage*100:.2f}%')
    else:
        raise NotImplementedError

    if show_progress:
        my_bar.empty()
    return vectorstore

@retry(stop=stop_after_attempt(5), wait=wait_exponential(multiplier=1,max=60))
def upsert_docs_db(vectorstore,
                         chunk_batch: List[Document],
                         chunk_batch_ids: List[str]):
    """
    Upserts a batch of documents into a vector database. The lancghain call is identical between Pinecone and ChromaDB.
    This function handles issues with hosted database upserts or when using hugging face or other endpoint services which are less stable.

    Args:
        vectorstore (VectorStore): The VectorStore object representing the Pinecone or ChromaDB.
        chunk_batch (List[Document]): A list of Document objects representing the batch of documents to be upserted.
        chunk_batch_ids (List[str]): A list of strings representing the IDs of the documents in the batch.

    Returns:
        VectorStore: The updated VectorStore object after upserting the documents.
    """
    vectorstore.add_documents(documents=chunk_batch,
                              ids=chunk_batch_ids)
    return vectorstore
def upsert_docs(index_type: str, 
                index_name: str,
                vectorstore: any, 
                chunker: dict, 
                batch_size: int = 50, 
                show_progress: bool = False,
                local_db_path: str = '.'):
    """
    Upserts documents into the specified index.

    Args:
        index_type (str): The type of index to upsert the documents into.
        index_name (str): The name of the index.
        vectorstore (any): The vectorstore object to add documents to.
        chunker (dict): The chunker dictionary containing the documents to upsert.
        batch_size (int, optional): The batch size for upserting documents.
        show_progress (bool, optional): Whether to show progress during the upsert process.
        local_db_path (str, optional): The local path to the database folder.

    Returns:
        tuple: A tuple containing the updated vectorstore and retriever objects.
    """
    if show_progress:
        progress_text = "Upsert in progress..."
        my_bar = st.progress(0, text=progress_text)
    if chunker['rag'] == 'Standard':
        if index_type == "Pinecone" or index_type == "ChromaDB":
            for i in range(0, len(chunker['chunks']), batch_size):
                chunk_batch = chunker['chunks'][i:i + batch_size]
                chunk_batch_ids = [_stable_hash_meta(chunk.metadata) for chunk in chunk_batch]   # add ID which is the hash of metadata
                
                vectorstore = upsert_docs_db(vectorstore,
                                             chunk_batch, chunk_batch_ids)
                    
                if show_progress:
                    progress_percentage = i / len(chunker['chunks'])
                    my_bar.progress(progress_percentage, text=f'{progress_text}{progress_percentage*100:.2f}%')
            retriever = vectorstore.as_retriever()
        elif index_type == "RAGatouille":
            # Create an index from the vectorstore.
            # This will default to split documents into 256 tokens each.
            vectorstore.index(
                collection=[chunk.page_content for chunk in chunker['chunks']],
                document_ids=[_stable_hash_meta(chunk.metadata) for chunk in chunker['chunks']],
                index_name=index_name,
                overwrite_index=True,
                split_documents=True,
                document_metadatas=[chunk.metadata for chunk in chunker['chunks']]
            )

            retriever = vectorstore.as_langchain_retriever()
        else:
            raise NotImplementedError
    elif chunker['rag'] == 'Parent-Child':
        if index_type == "Pinecone" or index_type == "ChromaDB":
            lfs_path = Path(local_db_path).resolve() / 'local_file_store' / index_name
            store = LocalFileStore(lfs_path)
            
            id_key = "doc_id"
            retriever = MultiVectorRetriever(vectorstore=vectorstore, byte_store=store, id_key=id_key)

            for i in range(0, len(chunker['chunks']), batch_size):
                chunk_batch = chunker['chunks'][i:i + batch_size]
                chunk_batch_ids = [_stable_hash_meta(chunk.metadata) for chunk in chunk_batch]   # add ID which is the hash of metadata
                
                retriever.vectorstore = upsert_docs_db(retriever.vectorstore,
                                                       chunk_batch, chunk_batch_ids)
                
                if show_progress:
                    progress_percentage = i / len(chunker['chunks'])
                    my_bar.progress(progress_percentage, text=f'{progress_text}{progress_percentage*100:.2f}%')
            
            # Index parent docs all at once
            retriever.docstore.mset(list(zip(chunker['pages']['doc_ids'], chunker['pages']['parent_chunks'])))
        elif index_type == "RAGatouille":
            raise Exception('RAGAtouille only supports standard RAG.')
        else:
            raise NotImplementedError
    elif chunker['rag'] == 'Summary':
        if index_type == "Pinecone" or index_type == "ChromaDB":
            lfs_path = Path(local_db_path).resolve() / 'local_file_store' / index_name
            store = LocalFileStore(lfs_path)
            
            id_key = "doc_id"
            retriever = MultiVectorRetriever(vectorstore=vectorstore, byte_store=store, id_key=id_key)

            for i in range(0, len(chunker['summaries']), batch_size):
                chunk_batch = chunker['summaries'][i:i + batch_size]
                chunk_batch_ids = [_stable_hash_meta(chunk.metadata) for chunk in chunk_batch]   # add ID which is the hash of metadata
                
                retriever.vectorstore = upsert_docs_db(retriever.vectorstore,
                                                       chunk_batch, chunk_batch_ids)

                if show_progress:
                    progress_percentage = i / len(chunker['summaries'])
                    my_bar.progress(progress_percentage, text=f'{progress_text}{progress_percentage*100:.2f}%')
            
            # Index parent docs all at once
            retriever.docstore.mset(list(zip(chunker['pages']['doc_ids'], chunker['pages']['docs'])))
        elif index_type == "RAGatouille":
            raise Exception('RAGAtouille only supports standard RAG.')
        else:
            raise NotImplementedError
    else:
        raise NotImplementedError
    if show_progress:
        my_bar.empty()
    return vectorstore, retriever
def delete_index(index_type: str, 
                 index_name: str, 
                 rag_type: str,
                 local_db_path: str = '.'):
    """
    Deletes an index based on the specified index type.

    Args:
        index_type (str): The type of index to delete. Valid values are "Pinecone", "ChromaDB", or "RAGatouille".
        index_name (str): The name of the index to delete.
        rag_type (str): The type of RAG (RAGatouille) to delete. Valid values are "Parent-Child" or "Summary".
        local_db_path (str, optional): The path to the local database.

    Raises:
        NotImplementedError: If the index_type is not supported.

    """
    if index_type == "Pinecone":
        pc = pinecone_client(api_key=os.getenv('PINECONE_API_KEY'))
        try:
            pc.describe_index(index_name)
            pc.delete_index(index_name)
        except Exception as e:
            pass
        if rag_type == 'Parent-Child' or rag_type == 'Summary':
            try:
                shutil.rmtree(Path(local_db_path).resolve() / 'local_file_store' / index_name)
            except Exception as e:
                pass    # No need to do anything if it doesn't exist
    elif index_type == "ChromaDB":  
        try:
            persistent_client = chromadb.PersistentClient(path=os.path.join(local_db_path,'chromadb'))
            # indices = persistent_client.list_collections()
            # for idx in indices:
            #     if index_name in idx.name:
                    # persistent_client.delete_collection(name=idx.name)
            persistent_client.delete_collection(name=index_name)
        except Exception as e:
            pass
        # Delete local file store if they exist
        if rag_type == 'Parent-Child' or rag_type == 'Summary':
            try:
                shutil.rmtree(Path(local_db_path).resolve() / 'local_file_store' / index_name)
            except Exception as e:
                pass    # No need to do anything if it doesn't exist
    elif index_type == "RAGatouille":
        try:
            ragatouille_path = os.path.join(local_db_path, '.ragatouille/colbert/indexes', index_name)
            shutil.rmtree(ragatouille_path)
        except Exception as e:
            pass
    else:
        raise NotImplementedError
def _sanitize_raw_page_data(page):
    """
    Sanitizes the raw page data by removing unnecessary information and checking for meaningful content.
    If pages are merged, this must happen before the merging occurs.

    Args:
        page (Page): The raw page data to be sanitized.

    Returns:
        Page or None: The sanitized page data if it contains meaningful content, otherwise None.
    """
    
    # Yank out some things you'll never care about
    page.metadata['source'] = os.path.basename(page.metadata['source'])   # Strip path
    page.metadata['page'] = int(page.metadata['page']) + 1   # Pages are 0 based, update
    page.page_content = re.sub(r"(\w+)-\n(\w+)", r"\1\2", page.page_content)   # Merge hyphenated words
    page.page_content = re.sub(r"(?<!\n\s)\n(?!\s\n)", " ", page.page_content.strip())  # Fix newlines in the middle of sentences
    page.page_content = re.sub(r"\n\s*\n", "\n\n", page.page_content)   # Remove multiple newlines

    # Test if there is meaningful content
    text = page.page_content
    num_words = len(text.split())
    if len(text) == 0:
        return None
    alphanumeric_pct = sum(c.isalnum() for c in text) / len(text)
    if num_words < 5 or alphanumeric_pct < 0.3:
        return None
    else:
        return page
def _embedding_size(embedding_family:Union[OpenAIEmbeddings,
                                           VoyageAIEmbeddings,
                                           HuggingFaceInferenceAPIEmbeddings]):
    """
    Returns the size of the embedding for a given embedding model.

    Args:
        embedding_family (object): The embedding model to get the size for.

    Returns:
        int: The size of the embedding.

    Raises:
        NotImplementedError: If the embedding model is not supported.
    """
    # https://platform.openai.com/docs/models/embeddings
    if isinstance(embedding_family,OpenAIEmbeddings):
        name=embedding_family.model
        if name=="text-embedding-ada-002":
            return 1536
        elif name=="text-embedding-3-small":
            return 1536
        elif name=="text-embedding-3-large":
            return 3072
        else:
            raise NotImplementedError(f"The embedding model '{name}' is not available in config.json")
    # https://docs.voyageai.com/embeddings/
    elif isinstance(embedding_family,VoyageAIEmbeddings):
        name=embedding_family.model
        if name=="voyage-2":
            return 1024 
        elif name=="voyage-large-2":
            return 1536
        elif name=="voyage-large-2-instruct":
            return 1024
        else:
            raise NotImplementedError(f"The embedding model '{name}' is not available in config.json")
    # See model pages for embedding sizes
    elif isinstance(embedding_family,HuggingFaceInferenceAPIEmbeddings):
        name=embedding_family.model_name
        if name=="sentence-transformers/all-MiniLM-L6-v2":
            return 384
        elif name=="mixedbread-ai/mxbai-embed-large-v1":
            return 1024
        else:
            raise NotImplementedError(f"The embedding model '{name}' is not available in config.json")
    else:
        raise NotImplementedError(f"The embedding family '{embedding_family}' is not available in config.json")
def _stable_hash_meta(metadata: dict):
    """
    Calculates the stable hash of the given metadata dictionary.

    Args:
        metadata (dict): The dictionary containing the metadata.

    Returns:
        str: The stable hash of the metadata.

    """
    return hashlib.sha1(json.dumps(metadata, sort_keys=True).encode()).hexdigest()
def db_name(index_type:str,rag_type:str,index_name:str,model_name:bool=None,check:bool=True):
    """
    Generates a modified name based on the given parameters.

    Args:
        index_type (str): The type of index.
        rag_type (str): The type of RAG.
        index_name (str): The name of the index.
        model_name (bool, optional): The name of the model. Defaults to None.
        check (bool, optional): Whether to check the validity of the name. Defaults to True.

    Returns:
        str: The modified index name.
        
    Raises:
        ValueError: If the Pinecone index name is longer than 45 characters.
        ValueError: If the ChromaDB collection name is longer than 63 characters.
        ValueError: If the ChromaDB collection name does not start and end with an alphanumeric character.
        ValueError: If the ChromaDB collection name contains characters other than alphanumeric, underscores, or hyphens.
        ValueError: If the ChromaDB collection name contains two consecutive periods.
    """
    # Modify name if it's an advanded RAG type
    if rag_type == 'Parent-Child':
        index_name = index_name + "-parent-child"
    elif rag_type == 'Summary':
        model_name_temp=model_name.replace(".", "-").replace("/", "-").lower()
        model_name_temp = model_name_temp.split('/')[-1]    # Just get the model name, not org
        model_name_temp = model_name_temp[:3] + model_name_temp[-3:]    # First and last 3 characters
        
        model_name_temp=model_name_temp+"-"+_stable_hash_meta({"model_name":model_name})[:4]

        index_name = index_name + "-" + model_name_temp + "-summary"
    else:
        index_name = index_name
    
    # Check if the name is valid
    if check:
        if index_type=="Pinecone":
            if len(index_name) > 45:
                raise ValueError(f"The Pinecone index name must be less than 45 characters. Entry: {index_name}")
            else:
                return index_name
        elif index_type=="ChromaDB":
            if len(index_name) > 63:
                raise ValueError(f"The ChromaDB collection name must be less than 63 characters. Entry: {index_name}")
            if not index_name[0].isalnum() or not index_name[-1].isalnum():
                raise ValueError(f"The ChromaDB collection name must start and end with an alphanumeric character. Entry: {index_name}")
            if not re.match(r"^[a-zA-Z0-9_-]+$", index_name):
                raise ValueError(f"The ChromaDB collection name can only contain alphanumeric characters, underscores, or hyphens. Entry: {index_name}")
            if ".." in index_name:
                raise ValueError(f"The ChromaDB collection name cannot contain two consecutive periods. Entry: {index_name}")
            else:
                return index_name
def get_or_create_spotlight_viewer(df:pd.DataFrame,port:int=9000):
    """
    Get or create a Spotlight viewer for the given DataFrame.

    Args:
        df (pd.DataFrame): The DataFrame to display in the viewer.
        port (int, optional): The port number to use for the viewer. Defaults to 9000.

    Returns:
        spotlight.viewer: The existing or newly created Spotlight viewer.
    """
    # TODO if you try to close spotlight and reuse a port, you get an error that the port is unavailable
    # TODO The viewer does not work properly with merged documents, it does not show the source column properly
    viewers = spotlight.viewers()
    if viewers:
        for viewer in viewers[:-1]:
            viewer.close()
        existing_viewer=spotlight.viewers()[-1]
        return existing_viewer
    
    new_viewer = spotlight.show(dataset=df,
                                wait='auto',
                                port=port,
                                no_ssl=True)

    return new_viewer
def get_docs_questions_df(
        docs_db_directory: Path,
        docs_db_collection: str,
        questions_db_directory: Path,
        questions_db_collection: str,
        query_model:object):
    """
    Retrieves and combines documents and questions dataframes.

    Args:
        docs_db_directory (Path): The directory of the documents database.
        docs_db_collection (str): The name of the documents database collection.
        questions_db_directory (Path): The directory of the questions database.
        questions_db_collection (str): The name of the questions database collection.
        query_model (object): The query model object.

    Returns:
        pd.DataFrame: The combined dataframe containing documents and questions data.
    """
    # TODO there's definitely a way to not have to pass query_model, since it should be possible to pull from the db, try to remove this in future versions
    # TODO extend this functionality to Pinecone

    # Check if there exists a query database
    chroma_collections = [collection.name for collection in admin.show_chroma_collections(format=False)['message']]
    matching_collection = [collection for collection in chroma_collections if collection == questions_db_collection]
    if len(matching_collection) > 1:
        raise Exception('Matching collection not found or multiple matching collections found.')
    try:
        if not matching_collection:
            raise Exception('Query database not found. Please create a query database using the Chatbot page and a selected index.')
    except Exception as e:
        st.warning(f"{e}")
        st.stop()
    st.markdown(f"Query database found: {questions_db_collection}")

    docs_df = get_docs_df('ChromaDB',docs_db_directory, docs_db_collection, query_model)
    docs_df["type"] = "doc"
    questions_df = get_questions_df(questions_db_directory, questions_db_collection, query_model)
    questions_df["type"] = "question"

    questions_df["num_sources"] = questions_df["sources"].apply(len)
    questions_df["first_source"] = questions_df["sources"].apply(
        lambda x: next(iter(x), None)
    )

    if len(questions_df):
        docs_df["used_by_questions"] = docs_df["id"].apply(
            lambda doc_id: questions_df[
                questions_df["sources"].apply(lambda sources: doc_id in sources)
            ]["id"].tolist()
        )
    else:
        docs_df["used_by_questions"] = [[] for _ in range(len(docs_df))]
    docs_df["used_by_num_questions"] = docs_df["used_by_questions"].apply(len)
    docs_df["used_by_question_first"] = docs_df["used_by_questions"].apply(
        lambda x: next(iter(x), None)
    )

    df = pd.concat([docs_df, questions_df], ignore_index=True)
    return df
def get_docs_df(index_type: str, local_db_path: Path, index_name: str, query_model: object):
    """
    Retrieves documents from a database and returns them as a pandas DataFrame.

    Args:
        index_type (str): The type of index to use (e.g., "Pinecone", "ChromaDB").
        local_db_path (Path): The local path to the Chroma database.
        index_name (str): The name of the collection in the Chroma database.
        query_model (object): The embedding function used for querying the database.

    Returns:
        pd.DataFrame: A DataFrame containing the retrieved documents, along with their metadata and embeddings.
            The DataFrame has the following columns:
            - id: The ID of the document.
            - source: The source of the document.
            - page: The page number of the document (default: -1 if not available).
            - document: The content of the document.
            - embedding: The embedding of the document.
    """
    # Find rag_type
    if index_name.endswith('-parent-child'):
        rag_type = 'Parent-Child'
    elif index_name.endswith('-summary'):
        rag_type = 'Summary'
    else:
        rag_type = 'Standard'

    if index_type=='ChromaDB':
        persistent_client = chromadb.PersistentClient(path=os.path.join(local_db_path,'chromadb'))            
        vectorstore = Chroma(client=persistent_client,
                                collection_name=index_name,
                                embedding_function=query_model) 
        response = vectorstore.get(include=["metadatas", "documents", "embeddings"])
        
        df_out=pd.DataFrame(
            {
                "id": response["ids"],
                "source": [metadata.get("source") for metadata in response["metadatas"]],
                "page": [metadata.get("page", -1) for metadata in response["metadatas"]],
                "metadata": response["metadatas"],
                "document": response["documents"],
                "embedding": response["embeddings"],
            })  

    elif index_type=='Pinecone':
        # Connect to index
        pc = pinecone_client(api_key=os.getenv('PINECONE_API_KEY'))
        index = pc.Index(index_name) 
        # Obtain list of ids
        ids=[]
        for id in index.list():
            ids.extend(id)
        # Fetch vectors from IDs in chunks (to not overload the API)
        docs=[]
        chunk_size=200  # 200 doesn't error our
        for i in range(0, len(ids), chunk_size):
            vector=index.fetch(ids[i:i+chunk_size])['vectors']
            vector_data = []
            for key, value in vector.items():
                vector_data.append(value)
            docs.extend(vector_data)

        df_out=pd.DataFrame(
            {
                "id": ids,
                "source": [data['metadata']['source'] for data in docs],
                "page": [data['metadata']['page'] for data in docs],
                "metadata": [{'page':data['metadata']['page'],'source':data['metadata']['source']} for data in docs],
                "document": [data['metadata']['page_content'] for data in docs],
                "embedding": [data['values'] for data in docs],
            })  
        # raise NotImplementedError('Only ChromaDB is supported for now')

    # Add parent-doc or original-doc
    if rag_type!='Standard':
        json_data_list = []
        for i, row in df_out.iterrows():
            doc_id = row['metadata']['doc_id']
            file_path = os.path.join(os.getenv('LOCAL_DB_PATH'),'local_file_store',index_name,f"{doc_id}")
            with open(file_path, "r") as f:
                json_data = json.load(f)
            json_data=json_data['kwargs']['page_content']
            json_data_list.append(json_data)
        if rag_type=='Parent-Child':
            df_out['parent-doc'] = json_data_list
        elif rag_type=='Summary':
            df_out['original-doc'] = json_data_list

    return df_out
def get_questions_df(local_db_path: Path, index_name: str, query_model: object):
    """
    Retrieves questions and related information from a Chroma database and returns them as a pandas DataFrame.

    Args:
        local_db_path (Path): The local path to the Chroma database.
        index_name (str): The name of the collection in the Chroma database.
        query_model (object): The embedding function used for querying the Chroma database.

    Returns:
        pd.DataFrame: A DataFrame containing the retrieved questions, answers, sources, and embeddings.
    """
    persistent_client = chromadb.PersistentClient(path=os.path.join(local_db_path,'chromadb'))            
    vectorstore = Chroma(client=persistent_client,
                            collection_name=index_name,
                            embedding_function=query_model) 

    response = vectorstore.get(include=["metadatas", "documents", "embeddings"])
    return pd.DataFrame(
        {
            "id": response["ids"],
            "question": response["documents"],
            "answer": [metadata.get("answer") for metadata in response["metadatas"]],
            "sources": [
                metadata.get("sources").split(",") for metadata in response["metadatas"]
            ],
            "embedding": response["embeddings"],
        }
    )
def add_clusters(df,n_clusters:int,label_llm:object=None,doc_per_cluster:int=5):
    """
    Add clusters to a DataFrame based on the embeddings of its documents.

    Args:
        df (pd.DataFrame): The DataFrame containing the documents and their embeddings.
        n_clusters (int): The number of clusters to create.
        label_llm (object, optional): The label language model to use for generating cluster labels. Defaults to None.
        doc_per_cluster (int, optional): The number of documents to sample per cluster for label generation. Defaults to 5.

    Returns:
        pd.DataFrame: The DataFrame with the added cluster information.
    """
    matrix = np.vstack(df.embedding.values)
    kmeans = KMeans(n_clusters=n_clusters, init="k-means++", random_state=42)
    kmeans.fit(matrix)
    labels = kmeans.labels_
    df["Cluster"] = labels

    summary=[]
    if label_llm is not None:
        for i in range(n_clusters):
            print(f"Cluster {i} Theme:")
            chunks =  df[df.Cluster == i].document.sample(doc_per_cluster, random_state=42)
            llm_chain = CLUSTER_LABEL | label_llm
            summary.append(llm_chain.invoke(chunks))
            print(summary[-1].content)
        df["Cluster_Label"] = [summary[i].content for i in df["Cluster"]]
    
    return df
def export_to_hf_dataset(df: pd.DataFrame, dataset_name: str):
    """
    Export a pandas DataFrame to a Hugging Face dataset and push it to the Hugging Face Hub.

    Args:
        df (pd.DataFrame): The pandas DataFrame to be exported.
        dataset_name (str): The name of the dataset to be created in the Hugging Face Hub.

    Returns:
        None
    """
    hf_dataset = Dataset.from_pandas(df)
    hf_dataset.push_to_hub(dataset_name, token=os.getenv('HUGGINGFACEHUB_API_TOKEN'))
def archive_db(index_type:str,index_name:str,query_model:object,export_pickle:bool=False):
    df_temp=get_docs_df(index_type,os.getenv('LOCAL_DB_PATH'), index_name, query_model)

    if export_pickle:
        # Export pickle to db directory
        with open(os.path.join(os.getenv('LOCAL_DB_PATH'),f"archive_{index_type.lower()}_{index_name}.pickle"), "wb") as f:
            pickle.dump(df_temp, f)
    return df_temp

# TODO add function to unarchive db (create chroma db and associated local filestore from pickle file)


import data_processing
from prompts import TEST_QUERY_PROMPT

import inspect
import os
import json
import streamlit as st
from dotenv import load_dotenv,find_dotenv

import openai
from pinecone import Pinecone
import chromadb
from langchain_openai import ChatOpenAI

from langchain_openai import OpenAIEmbeddings
from langchain_voyageai import VoyageAIEmbeddings
from langchain_community.embeddings import HuggingFaceInferenceAPIEmbeddings
from ragatouille import RAGPretrainedModel

class SecretKeyException(Exception):
    """Exception raised for secret key related errors.

    Attributes:
        message -- explanation of the error
        id -- unique identifier for the error
    """

    def __init__(self, message, id):
        super().__init__(message)
        self.id = id
class DatabaseException(Exception):
    """Exception raised for database related errors.

    Attributes:
        message -- explanation of the error
        id -- unique identifier for the error
    """

    def __init__(self, message, id):
        super().__init__(message)
        self.id = id
def load_sidebar(config_file,
                 vector_database=False,
                 embeddings=False,
                 rag_type=False,
                 index_selected=False,
                 llm=False,
                 model_options=False,
                 secret_keys=False):
    """
    Loads the sidebar configuration for the chatbot.

    Args:
        config_file (str): The path to the configuration file.
        vector_database (bool, optional): Whether to include the vector database in the sidebar.
        embeddings (bool, optional): Whether to include embeddings in the sidebar.
        rag_type (bool, optional): Whether to include RAG type in the sidebar.
        index_selected (bool, optional): Whether to include index name in the sidebar.
        llm (bool, optional): Whether to include LLM in the sidebar.
        model_options (bool, optional): Whether to include model options in the sidebar.
        secret_keys (bool, optional): Whether to include secret keys in the sidebar.

    Returns:
        dict: The sidebar configuration.
    """
    sb_out={}
    with open(config_file, 'r') as f:
        config = json.load(f)
        databases = {db['name']: db for db in config['databases']}
        embeddings_list = {e['name']: e for e in config['embeddings']}
        llms  = {m['name']: m for m in config['llms']}
        rag_types= config['rag_types']

    # Set local db path
    if os.getenv('LOCAL_DB_PATH') is None or os.getenv('LOCAL_DB_PATH')=='':
        # This is the case where the .env file is not in the directory
        raise SecretKeyException('Local Database Path is required. Use an absolute path for local, or /data for hugging face spaces.','LOCAL_DB_PATH_MISSING')

    # Vector databases
    if vector_database:
        st.sidebar.title('Vector database')
        sb_out['index_type']=st.sidebar.selectbox('Index type', list(databases.keys()), index=0,help='Select the type of index to use.')

        if embeddings:
            # Embeddings
            st.sidebar.title('Embeddings',help='See embedding leaderboard here for performance overview: https://huggingface.co/spaces/mteb/leaderboard')
            if sb_out['index_type']=='RAGatouille':    # Default to selecting hugging face model for RAGatouille, otherwise select alternates
                sb_out['query_model']=st.sidebar.selectbox('Hugging face rag models', 
                                                        databases[sb_out['index_type']]['embedding_models'], 
                                                        index=0,
                                                        help="Models listed are compatible with the selected index type.")
                sb_out['embedding_name']=sb_out['query_model']
            else:
                sb_out['query_model']=st.sidebar.selectbox('Embedding model family', 
                                                        databases[sb_out['index_type']]['embedding_models'], 
                                                        index=0,
                                                        help="Model provider.")
                sb_out['embedding_name']=st.sidebar.selectbox('Embedding model', 
                                                        embeddings_list[sb_out['query_model']]['embedding_models'], 
                                                        index=0,
                                                        help="Models listed are compatible with the selected index type.")
                if sb_out['embedding_name']=="Dedicated Endpoint":
                    sb_out['embedding_hf_endpoint']=st.sidebar.text_input('Dedicated endpoint URL','',
                                                        help='See Hugging Face configuration for endpoint. https://huggingface.co/inference-endpoints/dedicated')
                else:
                    sb_out['embedding_hf_endpoint']=None
        if rag_type:
            # RAG Type
            st.sidebar.title('RAG Type')
            if sb_out['index_type']=='RAGatouille':
                sb_out['rag_type']=st.sidebar.selectbox('RAG type', ['Standard'], index=0,help='Only Standard is available for RAGatouille.')
            else:
                sb_out['rag_type']=st.sidebar.selectbox('RAG type', rag_types, index=0,help='Parent-Child is for parent-child RAG. Summary is for summarization RAG.')
                if sb_out['rag_type']=='Summary':
                    sb_out['rag_llm_source']=st.sidebar.selectbox('RAG LLM model', list(llms.keys()), index=0,help='Select the LLM model for RAG.')
                    if sb_out['rag_llm_source']=='OpenAI':
                        sb_out['rag_llm_model']=st.sidebar.selectbox('RAG OpenAI model', llms[sb_out['rag_llm_source']]['models'], index=0,help='Select the OpenAI model for RAG.')
                    if sb_out['rag_llm_source']=='Hugging Face':
                        sb_out['rag_llm_model']=st.sidebar.selectbox('RAG Hugging Face model', 
                                                                llms['Hugging Face']['models'], 
                                                                index=0,
                                                                help='Select the Hugging Face model for RAG.')
                        if sb_out['rag_llm_model']=="Dedicated Endpoint":
                            sb_out['rag_hf_endpoint']=st.sidebar.text_input('Dedicated endpoint URL','',
                                                                help='See Hugging Face configuration for endpoint. https://huggingface.co/inference-endpoints/dedicated')
                            sb_out['rag_hf_endpoint']=sb_out['rag_hf_endpoint']+'/v1/'
                        else:
                            sb_out['rag_hf_endpoint']='https://api-inference.huggingface.co/v1'
                    elif sb_out['rag_llm_source']=='LM Studio (local)':
                        sb_out['rag_llm_model']=st.sidebar.text_input('Local host URL',
                                                                'http://localhost:1234/v1',
                                                                help='See LM studio configuration for local host URL.')
                        st.sidebar.warning('You must load a model in LM studio first for this to work.')
        if index_selected:
            if embeddings and rag_type:
                # Index Name 
                st.sidebar.title('Index Selected')  
                name=[]
                # For each index type, list indices available for the base name
                if sb_out['index_type']=='ChromaDB':
                    indices=show_chroma_collections(format=False)
                    if indices['status']:
                        for index in indices['message']:
                            # Be compatible with embedding types already used. Pinecone only supports lowercase.
                            if index.name.startswith((sb_out['embedding_name'].replace('/', '-')).lower()):    
                                if not index.name.endswith('-queries'): # Don't list query database as selectable
                                    if sb_out['rag_type']=='Parent-Child':
                                        if index.name.endswith('-parent-child'):
                                            name.append(index.name)
                                    elif sb_out['rag_type']=='Summary':
                                        if index.name.endswith('-summary'):
                                            name.append(index.name)
                                    else:   # Stadard
                                        if not index.name.endswith('-parent-child') and not index.name.endswith('-summary'):
                                            name.append(index.name)
                        sb_out['index_selected']=st.sidebar.selectbox('Index selected',name,index=0,help='Select the index to use for the application.')
                    else:
                        st.sidebar.markdown('No collections found.',help='Check the status on Home.')
                elif sb_out['index_type']=='Pinecone':
                    indices=show_pinecone_indexes(format=False)
                    if indices['status']:
                        for index in indices['message']:
                            # if index['status']['state']=='Ready':
                            #     name.append(index['name'])
                            if index['name'].startswith((sb_out['embedding_name'].replace('/', '-')).lower()):    
                                if not index['name'].endswith('-queries'): # Don't list query database as selectable
                                    if sb_out['rag_type']=='Parent-Child':
                                        if index['name'].endswith('-parent-child'):
                                            name.append(index['name'])
                                    elif sb_out['rag_type']=='Summary':
                                        if index['name'].endswith('-summary'):
                                            name.append(index['name'])
                                    else:
                                        name.append(index['name'])
                        sb_out['index_selected']=st.sidebar.selectbox('Index selected',name,index=0,help='Select the index to use for the application.')
                elif sb_out['index_type']=='RAGatouille':
                    indices=show_ragatouille_indexes(format=False)
                    if len(indices)>0:
                        for index in indices['message']:
                            # Be compatible with embedding types already used. Pinecone only supports lowercase.
                            if index.startswith((sb_out['embedding_name'].replace('/', '-')).lower()):    
                                name.append(index)
                        sb_out['index_selected']=st.sidebar.selectbox('Index selected',name,index=0,help='Select the index to use for the application.')
                    else:
                        st.sidebar.markdown('No collections found.',help='Check the status on Home.')
                else:
                    raise NotImplementedError
                try:
                    if not name:
                        raise DatabaseException('No collections found for the selected index type/embedding. Create a new database, or select another index type/embedding.','NO_COMPATIBLE_COLLECTIONS')
                except DatabaseException as e:
                    st.warning(f"{e}")
                    st.stop()
            else:
                raise ValueError('Embeddings must be enabled to select an index name.')
        if llm:
            # LLM
            st.sidebar.title('LLM',help='See LLM leaderboard here for performance overview: https://huggingface.co/spaces/lmsys/chatbot-arena-leaderboard')
            sb_out['llm_source']=st.sidebar.selectbox('LLM model', list(llms.keys()), index=0,help='Select the LLM model for the application.')
            if sb_out['llm_source']=='OpenAI':
                sb_out['llm_model']=st.sidebar.selectbox('OpenAI model', llms[sb_out['llm_source']]['models'], index=0,help='Select the OpenAI model for the application.')
            elif sb_out['llm_source']=='Hugging Face':
                sb_out['llm_model']=st.sidebar.selectbox('Hugging Face model', 
                                                        llms['Hugging Face']['models'], 
                                                        index=0,
                                                        help='Select the Hugging Face model for the application.')
                sb_out['hf_endpoint']='https://api-inference.huggingface.co/v1'

                if sb_out['llm_model']=="Dedicated Endpoint":
                    sb_out['hf_endpoint']=st.sidebar.text_input('Dedicated endpoint URL','',
                                                        help='See Hugging Face configuration for endpoint. https://huggingface.co/inference-endpoints/dedicated')
                    sb_out['hf_endpoint']=sb_out['hf_endpoint']+'/v1/'
                else:
                    sb_out['hf_endpoint']='https://api-inference.huggingface.co/v1'
            elif sb_out['llm_source']=='LM Studio (local)':
                sb_out['llm_model']=st.sidebar.text_input('Local host URL',
                                                        'http://localhost:1234/v1',
                                                        help='See LM studio configuration for local host URL.')
                st.sidebar.warning('You must load a model in LM studio first for this to work.')
        if model_options:
            # Add input fields in the sidebar
            st.sidebar.title('LLM Options')
            temperature = st.sidebar.slider('Temperature', min_value=0.0, max_value=2.0, value=0.1, step=0.1,help='Temperature for LLM.')
            output_level = st.sidebar.number_input('Max output tokens', min_value=50, step=10, value=1000,
                                                help='Max output tokens for LLM. Concise: 50, Verbose: 1000. Limit depends on model.')
            # Set different options for if ragatouille is used, since it has fewer parameters to select
            if 'index_type' in sb_out:
                st.sidebar.title('Retrieval Options')
                k = st.sidebar.number_input('Number of items per prompt', min_value=1, step=1, value=4,help='Number of items to retrieve per query.')
                if sb_out['index_type']!='RAGatouille':
                    search_type = st.sidebar.selectbox('Search Type', ['similarity', 'mmr'], index=0,help='Select the search type for the application.')
                    sb_out['model_options']={'output_level':output_level,
                                            'k':k,
                                            'search_type':search_type,
                                            'temperature':temperature}
                else:
                    sb_out['model_options']={'output_level':output_level,
                                             'k':k,
                                            'search_type':None,
                                            'temperature':temperature}
    else:
        if embeddings or rag_type or index_selected or llm or model_options:
            # Must have vector database for any of this functionality.
            raise ValueError('Vector database must be enabled to use these options.')

    # Secret keys, which does not rely on vector_database
    if secret_keys:
        sb_out['keys']={}
        # Add a section for secret keys
        st.sidebar.title('Secret keys',help='See Home page under Connection Status for status of keys.')
        st.sidebar.markdown('If .env file is in directory, will use that first.')
        if 'llm_source' in sb_out and sb_out['llm_source'] == 'OpenAI':
            sb_out['keys']['OPENAI_API_KEY'] = st.sidebar.text_input('OpenAI API Key', type='password',help='OpenAI API Key: https://platform.openai.com/api-keys')
        elif 'query_model' in sb_out and sb_out['query_model'] == 'OpenAI':
            sb_out['keys']['OPENAI_API_KEY'] = st.sidebar.text_input('OpenAI API Key', type='password',help='OpenAI API Key: https://platform.openai.com/api-keys')
        if 'llm_source' in sb_out and sb_out['llm_source']=='Hugging Face':
            sb_out['keys']['HUGGINGFACEHUB_API_TOKEN'] = st.sidebar.text_input('Hugging Face API Key', type='password',help='Hugging Face API Key: https://huggingface.co/settings/tokens')
        if 'query_model' in sb_out and sb_out['query_model']=='Voyage':
            sb_out['keys']['VOYAGE_API_KEY'] = st.sidebar.text_input('Voyage API Key', type='password',help='Voyage API Key: https://dash.voyageai.com/api-keys')
        if 'index_type' in sb_out and sb_out['index_type']=='Pinecone':
            sb_out['keys']['PINECONE_API_KEY']=st.sidebar.text_input('Pinecone API Key',type='password',help='Pinecone API Key: https://www.pinecone.io/')
            
    return sb_out
def set_secrets(sb):
    """
    Sets the secrets for various API keys by retrieving them from the environment variables or the sidebar.

    Args:
        sb (dict): The sidebar data containing the API keys.

    Returns:
        dict: A dictionary containing the set API keys.

    Raises:
        SecretKeyException: If any of the required API keys are missing.
    """
    secrets={}

    secrets['OPENAI_API_KEY'] = os.getenv('OPENAI_API_KEY')
    if not secrets['OPENAI_API_KEY'] and 'keys' in sb and 'OPENAI_API_KEY' in sb['keys']:
        secrets['OPENAI_API_KEY'] = sb['keys']['OPENAI_API_KEY']
        os.environ['OPENAI_API_KEY'] = secrets['OPENAI_API_KEY']
        if os.environ['OPENAI_API_KEY']=='':
            raise SecretKeyException('OpenAI API Key is required.','OPENAI_API_KEY_MISSING')
    openai.api_key = secrets['OPENAI_API_KEY']

    secrets['VOYAGE_API_KEY'] = os.getenv('VOYAGE_API_KEY')
    if not secrets['VOYAGE_API_KEY'] and 'keys' in sb and 'VOYAGE_API_KEY' in sb['keys']:
        secrets['VOYAGE_API_KEY'] = sb['keys']['VOYAGE_API_KEY']
        os.environ['VOYAGE_API_KEY'] = secrets['VOYAGE_API_KEY']
        if os.environ['VOYAGE_API_KEY']=='':
            raise SecretKeyException('Voyage API Key is required.','VOYAGE_API_KEY_MISSING')

    secrets['PINECONE_API_KEY'] = os.getenv('PINECONE_API_KEY')
    if not secrets['PINECONE_API_KEY'] and 'keys' in sb and 'PINECONE_API_KEY' in sb['keys']:
        secrets['PINECONE_API_KEY'] = sb['keys']['PINECONE_API_KEY']
        os.environ['PINECONE_API_KEY'] = secrets['PINECONE_API_KEY']
        if os.environ['PINECONE_API_KEY']=='':
            raise SecretKeyException('Pinecone API Key is required.','PINECONE_API_KEY_MISSING')

    secrets['HUGGINGFACEHUB_API_TOKEN'] = os.getenv('HUGGINGFACEHUB_API_TOKEN')
    if not secrets['HUGGINGFACEHUB_API_TOKEN'] and 'keys' in sb and 'HUGGINGFACEHUB_API_TOKEN' in sb['keys']:
        secrets['HUGGINGFACEHUB_API_TOKEN'] = sb['keys']['HUGGINGFACEHUB_API_TOKEN']
        os.environ['HUGGINGFACEHUB_API_TOKEN'] = secrets['HUGGINGFACEHUB_API_TOKEN']
        if os.environ['HUGGINGFACEHUB_API_TOKEN']=='':
            raise SecretKeyException('Hugging Face API Key is required.','HUGGINGFACE_API_KEY_MISSING')
    return secrets
def test_key_status():
    """
    Check the status of various API keys based on environment variables.

    Returns:
        dict: A dictionary containing the status of each API key.
    """
    key_status = {}
    # OpenAI
    if os.getenv('OPENAI_API_KEY') is None or os.getenv('OPENAI_API_KEY') == '':
        key_status['OpenAI API Key'] = {'status': False}
    else:
        key_status['OpenAI API Key'] = {'status': True}

    # Voyage
    if os.getenv('VOYAGE_API_KEY') is None or os.getenv('VOYAGE_API_KEY') == '':
        key_status['Voyage API Key'] = {'status': False}
    else:
        key_status['Voyage API Key'] = {'status': True}

    # Pinecone
    if os.getenv('PINECONE_API_KEY') is None or os.getenv('PINECONE_API_KEY') == '':
        key_status['Pinecone API Key'] = {'status': False}
    else:
        key_status['PINECONE_API_KEY'] = {'status': True}
    
    # Hugging Face
    if os.getenv('HUGGINGFACEHUB_API_TOKEN') is None or os.getenv('HUGGINGFACEHUB_API_TOKEN') == '':
        key_status['Hugging Face API Key'] = {'status': False}
    else:
        key_status['Hugging Face API Key'] = {'status': True}

    return _format_key_status(key_status)
def set_llm(sb, secrets, type='prompt'):
    """
    Sets up and returns a language model (LLM) based on the provided parameters.

    Args:
        sb (dict): The configuration settings for the chatbot.
        secrets (dict): The secret keys and tokens required for authentication.
        type (str, optional): The type of LLM to set up.

    Returns:
        ChatOpenAI: The configured language model.

    Raises:
        ValueError: If an invalid LLM source is specified.

    """
    if type == 'prompt':  # use for prompting in chat applications
        if sb['llm_source'] == 'OpenAI':
            llm = ChatOpenAI(model_name=sb['llm_model'],
                             temperature=sb['model_options']['temperature'],
                             openai_api_key=secrets['OPENAI_API_KEY'],
                             max_tokens=sb['model_options']['output_level'],
                             tags=[sb['llm_model']])
        elif sb['llm_source'] == 'Hugging Face':
            llm = ChatOpenAI(base_url=sb['hf_endpoint'],
                             model=sb['llm_model'],
                             api_key=secrets['HUGGINGFACEHUB_API_TOKEN'],
                             temperature=sb['model_options']['temperature'],
                             max_tokens=sb['model_options']['output_level'],
                             tags=[sb['llm_model']])
        elif sb['llm_source'] == 'LM Studio (local)':
            # base_url takes local configuration from lm studio, no api key required.
            llm = ChatOpenAI(base_url=sb['llm_model'],
                             temperature=sb['model_options']['temperature'],
                             max_tokens=sb['model_options']['output_level'])
        else:
            raise ValueError("Invalid LLM source specified.")
    elif type == 'rag':   # use for RAG application (summary)
        if sb['rag_llm_source'] == 'OpenAI':
            llm = ChatOpenAI(model_name=sb['rag_llm_model'],
                             openai_api_key=secrets['OPENAI_API_KEY'],
                             temperature=sb['model_options']['temperature'],
                             max_tokens=sb['model_options']['output_level'],
                             tags=[sb['rag_llm_model']])
        elif sb['rag_llm_source'] == 'Hugging Face':
            llm = ChatOpenAI(base_url=sb['rag_hf_endpoint'],
                             model=sb['rag_llm_model'],
                             temperature=sb['model_options']['temperature'],
                             max_tokens=sb['model_options']['output_level'],
                             api_key=secrets['HUGGINGFACEHUB_API_TOKEN'],
                             tags=[sb['rag_llm_model']])
        elif sb['rag_llm_source'] == 'LM Studio (local)':
            # base_url takes local configuration from lm studio, no api key required.
            llm = ChatOpenAI(base_url=sb['rag_llm_model'],
                             model_name='lm_studio')
        else:
            raise ValueError("Invalid LLM source specified.")
    return llm
def get_query_model(sb, secrets):
    """
    Returns the query model based on the provided parameters.

    Args:
        sb (dict): A dictionary containing the parameters for the query model.
        secrets (dict): A dictionary containing the API keys for different query models.

    Returns:
        query_model: The selected query model based on the provided parameters.

    Raises:
        NotImplementedError: If the query model is not recognized.
    """
    if sb['index_type'] == 'RAGatouille':
        query_model = RAGPretrainedModel.from_pretrained(sb['embedding_name'],
                                                         index_root=os.path.join(os.getenv('LOCAL_DB_PATH'),'.ragatouille'))
    elif sb['query_model'] == 'OpenAI':
        query_model = OpenAIEmbeddings(model=sb['embedding_name'], openai_api_key=secrets['OPENAI_API_KEY'])
    elif sb['query_model'] == 'Voyage':
        query_model = VoyageAIEmbeddings(model=sb['embedding_name'], voyage_api_key=secrets['VOYAGE_API_KEY'], truncation=False)
    elif sb['query_model'] == 'Hugging Face':
        if sb['embedding_hf_endpoint']:
            query_model = HuggingFaceInferenceAPIEmbeddings(model_name=sb['embedding_name'], 
                                                            api_url=sb['embedding_hf_endpoint'],
                                                            api_key=secrets['HUGGINGFACEHUB_API_TOKEN'])
        else:
            query_model = HuggingFaceInferenceAPIEmbeddings(model_name=sb['embedding_name'], 
                                                            api_key=secrets['HUGGINGFACEHUB_API_TOKEN'])
    else:
        raise NotImplementedError('Query model not recognized.')
    return query_model
def show_pinecone_indexes(format=True):
    """
    Retrieves the list of Pinecone indexes and their status.
    LOCAL_DB_PATH environment variable used to pass the local database path.

    Args:
        format (bool, optional): Specifies whether to format the output for display.

    Returns:
        dict or str: If format is True, returns a formatted string representation of the Pinecone status.
                    If format is False, returns a dictionary containing the Pinecone status.

    """
    if os.getenv('PINECONE_API_KEY') is None or os.getenv('PINECONE_API_KEY')=='':
        pinecone_status = {'status': False, 'message': 'Pinecone API Key is not set.'}
    else:
        pc=Pinecone(api_key=os.getenv('PINECONE_API_KEY'))
        indexes=pc.list_indexes()
        if len(indexes)==0:
            pinecone_status = {'status': False, 'message': 'No indexes found'}
        else:
            pinecone_status = {'status': True, 'message': indexes}
    
    if format:
        return _format_pinecone_status(pinecone_status)
    else:
        return pinecone_status
def show_chroma_collections(format=True):
    """
    Retrieves the list of chroma collections from the local database.
    LOCAL_DB_PATH environment variable used to pass the local database path.

    Args:
        format (bool, optional): Specifies whether to format the output for display.

    Returns:
        dict or str: If format is True, returns a formatted string representation of the chroma status.
                    If format is False, returns a dictionary containing the chroma status.

    Raises:
        ValueError: If the chroma vector database needs to be reset.

    """
    if os.getenv('LOCAL_DB_PATH') is None or os.getenv('LOCAL_DB_PATH')=='':
        chroma_status = {'status': False, 'message': 'Local database path is not set.'}
    else:
        db_folder_path=os.getenv('LOCAL_DB_PATH')
        try:
            persistent_client = chromadb.PersistentClient(path=os.path.join(db_folder_path,'chromadb'))
        except:
            raise ValueError("Chroma vector database needs to be reset, or the database path is incorrect. Clear cache, or reset path. You may have specified a path which is read only or has no collections.")
        collections=persistent_client.list_collections()
        if len(collections)==0:
            chroma_status = {'status': False, 'message': 'No collections found'}
        else:   
            chroma_status = {'status': True, 'message': collections}
    if format:
        return _format_chroma_status(chroma_status)
    else:
        return chroma_status
def show_ragatouille_indexes(format=True):
    """
    Retrieves the list of ragatouille indexes.
    LOCAL_DB_PATH environment variable used to pass the local database path.

    Args:
        format (bool, optional): Specifies whether to format the indexes for display.

    Returns:
        dict or str: If format is True, returns a formatted string representation of the ragatouille status.
                    If format is False, returns a dictionary containing the ragatouille status.

    Raises:
        ValueError: If the ragatouille vector database needs to be reset.

    """
    if os.getenv('LOCAL_DB_PATH') is None or os.getenv('LOCAL_DB_PATH')=='':
        ragatouille_status = {'status': False, 'message': 'Local database path is not set.'}
    else:
        db_folder_path=os.getenv('LOCAL_DB_PATH')
        try:
            path=os.path.join(db_folder_path,'.ragatouille/colbert/indexes')
            indexes = []
            for item in os.listdir(path):
                item_path = os.path.join(path, item)
                if os.path.isdir(item_path):
                    indexes.append(item)
            if len(indexes)>0:
                ragatouille_status = {'status': True, 'message': indexes}
            else:
                ragatouille_status = {'status': False, 'message': 'No indexes found.'}  # if .ragatouille structure exists but is empty
        except:
            ragatouille_status = {'status': False, 'message': 'No indexes found.'}  # if .ragatouille structure does not exist yet

    if format:
        return _format_ragatouille_status(ragatouille_status)
    else:
        return ragatouille_status
def st_connection_status_expander(expanded: bool = True, delete_buttons: bool = False, set_secrets: bool = False):
    """
    Makes a Streamlit expander widget to display connection status information.
    LOCAL_DB_PATH environment variable used to pass the local database path.

    Args:
        expanded (bool, optional): Whether the expander is initially expanded or collapsed. Only intended with account access.
        delete_buttons (bool, optional): Whether to display delete buttons for Pinecone and Chroma DB indexes.
        set_secrets (bool, optional): Whether to set the secrets.
    """
    # TODO add capability to add dedicated endpoints for Hugging Face models
    with st.expander("Connection Status", expanded=expanded):
        # Set secrets and assign to environment variables
        if set_secrets:
            st.markdown("**Set API keys**:")
            keys={}
            # OPENAI_API_KEY
            keys['OPENAI_API_KEY'] = st.text_input('OpenAI API Key', type='password',help='OpenAI API Key: https://platform.openai.com/api-keys')
            if keys['OPENAI_API_KEY']!='':
                os.environ['OPENAI_API_KEY'] = keys['OPENAI_API_KEY'] 
            
            # VOYAGE_API_KEY
            keys['VOYAGE_API_KEY'] = st.text_input('Voyage API Key', type='password',help='Voyage API Key: https://dash.voyageai.com/api-keys')
            if keys['VOYAGE_API_KEY']!='':
                os.environ['VOYAGE_API_KEY'] = keys['VOYAGE_API_KEY']
            
            # PINECONE_API_KEY
            keys['PINECONE_API_KEY']=st.text_input('Pinecone API Key',type='password',help='Pinecone API Key: https://www.pinecone.io/')
            if keys['PINECONE_API_KEY']!='':
                os.environ['PINECONE_API_KEY'] = keys['PINECONE_API_KEY']

            # HUGGINGFACEHUB_API_TOKEN
            keys['HUGGINGFACEHUB_API_TOKEN'] = st.text_input('Hugging Face API Key', type='password',help='Hugging Face API Key: https://huggingface.co/settings/tokens')
            if keys['HUGGINGFACEHUB_API_TOKEN']!='':
                os.environ['HUGGINGFACEHUB_API_TOKEN'] = keys['HUGGINGFACEHUB_API_TOKEN']
            
            # LOCAL_DB_PATH
            keys['LOCAL_DB_PATH'] = st.text_input('Update Local Database Path',os.getenv('LOCAL_DB_PATH'),help='Path to local database (e.g. chroma)')
            if keys['LOCAL_DB_PATH']!='':
                os.environ['LOCAL_DB_PATH'] = keys['LOCAL_DB_PATH']

        if os.getenv('LOCAL_DB_PATH') is None:
            raise SecretKeyException('Local Database Path is required. Use an absolute path.','LOCAL_DB_PATH_MISSING')
        else:
            db_folder_path=os.getenv('LOCAL_DB_PATH')

        # Show key status
        st.markdown("**API key status** (Indicates status of local variable. It does not guarantee the key itself is correct):")
        st.markdown(test_key_status())

        # TODO the try statements below sometimes allow delete to be pressed but the index which does exist isn't deleted. Update so an error is thrown if the index is not deleted.
        # Pinecone
        st.markdown(show_pinecone_indexes())
        try:
            pinecone_indexes = [obj.name for obj in show_pinecone_indexes(format=False)['message']]
            if delete_buttons:
                pinecone_index_name = st.selectbox('Pinecone index to delete', pinecone_indexes)
                if st.button('Delete Pinecone index', help='This is permanent!'):
                    if pinecone_index_name.endswith("parent-child"):
                        rag_type = "Parent-Child"
                    elif pinecone_index_name.endswith("summary"):
                        rag_type = "Summary"
                    else:
                        rag_type = "Standard"
                    data_processing.delete_index('Pinecone', pinecone_index_name, rag_type, local_db_path=db_folder_path)
                    st.markdown(f"Index {pinecone_index_name} has been deleted.")
        except:
            pass

        # Chroma DB
        st.markdown(show_chroma_collections())
        try:
            chroma_db_collections = [obj.name for obj in show_chroma_collections(format=False)['message']]
            if delete_buttons:
                chroma_db_name = st.selectbox('Chroma database to delete', chroma_db_collections)
                if st.button('Delete Chroma database', help='This is permanent!'):
                    if chroma_db_name.endswith("parent-child"):
                        rag_type = "Parent-Child"
                    elif "-summary-" in chroma_db_name:
                        rag_type = "Summary"
                    else:
                        rag_type = "Standard"
                    data_processing.delete_index('ChromaDB', chroma_db_name, rag_type, local_db_path=db_folder_path)
                    st.markdown(f"Database {chroma_db_name} has been deleted.")
        except:
            pass
        
        # Ragatouille
        st.markdown(show_ragatouille_indexes())
        try:
            ragatouille_index_data=show_ragatouille_indexes(format=False)
            ragatouille_indexes = [obj for obj in ragatouille_index_data['message']]
            if delete_buttons and ragatouille_index_data['status']==True:
                ragatouille_name = st.selectbox('RAGatouille database to delete', ragatouille_indexes)
                if st.button('Delete RAGatouille database', help='This is permanent!'):
                    data_processing.delete_index('RAGatouille', ragatouille_name, "Standard", local_db_path=db_folder_path)
                    st.markdown(f"Index {ragatouille_name} has been deleted.")
        except:
            pass

        # Local database path
        st.markdown(f"Local database path: {os.environ['LOCAL_DB_PATH']}")
def st_setup_page(page_title: str, home_dir:str, config_file: str, sidebar_config: dict = None):
    """
    Sets up the Streamlit page with the given title and loads the sidebar configuration.

    Args:
        page_title (str): The title of the Streamlit page.
        home_dir (str): The path to the home directory.
        config_file (str): The path to the configuration file.
        sidebar_config (dict, optional): The sidebar configuration.

    Returns:
        tuple: A tuple containing the following:
            - paths (dict): A dictionary containing the following directory paths:
                - base_folder_path (str): The path to the root folder.
                - config_folder_path (str): The path to the config folder.
                - data_folder_path (str): The path to the data folder.
                - db_folder_path (str): The path to the database folder.
            - sb (dict): The sidebar configuration.
            - secrets (dict): A dictionary containing the set API keys.

    Raises:
        SecretKeyException: If there is an issue with the secret keys.

    """
    load_dotenv(find_dotenv(), override=True)

    base_folder_path = home_dir
    data_folder_path=os.path.join(base_folder_path, 'data')

    # Set the page title
    st.title(page_title)

    # Set local database
    # Only show the text input if no value has been entered yet
    if not os.environ.get('LOCAL_DB_PATH'):
        local_db_path_input = st.empty()  # Create a placeholder for the text input
        warn_db_path=st.warning('Local Database Path is required to initialize. Use an absolute path.')
        local_db_path = local_db_path_input.text_input('Update Local Database Path', help='Path to local database (e.g. chroma).')
        if local_db_path:
            os.environ['LOCAL_DB_PATH'] = local_db_path
        else:
            st.stop()
    if os.environ.get('LOCAL_DB_PATH'): # If a value has been entered, update the environment variable and clear the text input
        try:
            local_db_path_input.empty()  # This will remove the text input from the page if it exists
            warn_db_path.empty()
        except:
            pass    # If the text input has already been removed, do nothing

    # Load sidebar
    try:
        if sidebar_config is None:
            sb=load_sidebar(config_file)
        else:
            sb=load_sidebar(config_file,
                            **sidebar_config)
    except SecretKeyException as e:
        # If no .env file is found, set the local db path when the warning is raised.
        st.warning(f"{e}")
        st.stop()
    try:
        secrets=set_secrets(sb) # Take secrets from .env file first, otherwise from sidebar
    except SecretKeyException as e:
        st.warning(f"{e}")
        st.stop()

    # Set db folder path based on env variable
    db_folder_path=os.environ['LOCAL_DB_PATH']

    paths={'base_folder_path':base_folder_path,
           'config_folder_path':config_file,
           'data_folder_path':data_folder_path,
           'db_folder_path':db_folder_path}

    return paths,sb,secrets
def _format_key_status(key_status: str):
    """
    Formats the key status dictionary into a formatted string.

    Args:
        key_status (str): The dictionary containing the key status information.

    Returns:
        str: The formatted string representing the key status.
    """
    formatted_status = ""
    for key, value in key_status.items():
        status = value['status']
        if status:
            formatted_status += f"- {key}: :heavy_check_mark:\n"
        else:
            formatted_status += f"- {key}: :x:\n"
    return formatted_status
def _format_pinecone_status(pinecone_status):
    """
    Formats the Pinecone status into a markdown string.

    Args:
        pinecone_status (dict): The Pinecone status dictionary.

    Returns:
        str: The formatted markdown string.
    """
    index_description=''
    if pinecone_status['status']:
        for index in pinecone_status['message']:
            name = index['name']
            state = index['status']['state']
            status = ":heavy_check_mark:"
            index_description += f"- {name}: {state} ({status})\n"
        markdown_string = f"**Pinecone Indexes**\n{index_description}"
    else:
        message = pinecone_status['message']
        markdown_string = f"**Pinecone Indexes**\n- {message} :x:"
    return markdown_string
def _format_chroma_status(chroma_status):
    """
    Formats the chroma status dictionary into a markdown string.

    Args:
        chroma_status (dict): The chroma status dictionary containing the status and message.

    Returns:
        str: The formatted markdown string.
    """
    collection_description=''
    if chroma_status['status']:
        for index in chroma_status['message']:
            name = index.name
            status = ":heavy_check_mark:"
            collection_description += f"- {name}: ({status})\n"
        markdown_string = f"**ChromaDB Collections**\n{collection_description}"
    else:
        message = chroma_status['message']
        markdown_string = f"**ChromaDB Collections**\n- {message} :x:"
    return markdown_string
def _format_ragatouille_status(ragatouille_status):
    """
    Formats the ragatouille status for display.

    Args:
        ragatouille_status (dict): The ragatouille status dictionary.

    Returns:
        str: A formatted string representation of the ragatouille status.

    """
    index_description=''
    if ragatouille_status['status']:
        for index in ragatouille_status['message']:
            name = index
            status = ":heavy_check_mark:"
            index_description += f"- {name}: ({status})\n"
        markdown_string = f"**Ragatouille Indexes**\n{index_description}"
    else:
        message = ragatouille_status['message']
        markdown_string = f"**Ragatouille Indexes**\n- {message} :x:"
    return markdown_string


from langchain import hub
from langchain.prompts.prompt import PromptTemplate

# Prompts on the hub: https://smith.langchain.com/hub/dmueller?organizationId=45eb8917-7353-4296-978d-bb461fc45c65
CONDENSE_QUESTION_PROMPT = hub.pull("dmueller/ams-chatbot-qa-condense-history")
QA_PROMPT=hub.pull("dmueller/ams-chatbot-qa-retrieval")
QA_WSOURCES_PROMPT=hub.pull("dmueller/ams-chatbot-qa-retrieval-wsources")
QA_GENERATE_PROMPT=hub.pull("dmueller/generate_qa_prompt")
SUMMARIZE_TEXT=hub.pull("dmueller/summarize_text")
TEST_QUERY_PROMPT='What are examples of adhesives to use when potting motors for launch vehicle or spacecraft mechanisms?'
GENERATE_SIMILAR_QUESTIONS=hub.pull("dmueller/generate_similar_questions")
GENERATE_SIMILAR_QUESTIONS_W_CONTEXT=hub.pull("dmueller/generate_similar_questions_w_context")
CLUSTER_LABEL=hub.pull("dmueller/cluster-label")

# Prompts defined here only
DEFAULT_DOCUMENT_PROMPT = PromptTemplate.from_template(template="{page_content}")